import java.io.*;
import java.net.*;
import java.util.*;
import Parser;

/**
Euler proof mechanism.
  @version 27.035
  @author Jos De Roo
                                                                                                    */
public class Euler implements Serializable {
  Euler subject = null;                     // RDF subject
  String verb = null;                       // absoluteized verb
  Euler object = null;                      // RDF object
  String cverb = null;                      // compact verb
  int varid = -1;                           // variable id
  boolean bound = false;                    // variable binding state
  Euler ref = null;                         // to be dereferenced
  Euler near = null;                        // to construct a conjunction
  Euler far = this;                         // to remember last/root clause
  Hashtable ns = null;                      // to table namespaces
  Hashtable nsp = null;                     // to table namespace prefixes
  Hashtable syn = null;                     // to table synonyms
  Hashtable anon = null;                    // to table anonymous nodes
  Vector refl = null;                       // to flag NonIrreflexive properties
  Vector uniq = null;                       // to flag daml:UniqueProperty properties
  int varmax = 0;                           // maximum variable count
  String baseURI = null;                    // URI of base process
  Vector asserted = null;                   // to keep track of asserted documents

  static final String version = "27.035";
  static final long serialVersionUID = 1;
  static final String N3 = "<http://www.w3.org/2000/10/swap/log#>";
  static final String N3Truth = "<http://www.w3.org/2000/10/swap/log#Truth>";
  static final String N3Falsehood = "<http://www.w3.org/2000/10/swap/log#Falsehood>";
  static final String N3forAll = "<http://www.w3.org/2000/10/swap/log#forAll>";
  static final String N3forAllQN = "log:forAll";
  static final String N3forSome = "<http://www.w3.org/2000/10/swap/log#forSome>";
  static final String N3forSomeQN = "log:forSome";
  static final String N3implies = "<http://www.w3.org/2000/10/swap/log#implies>";
  static final String N3impliesQN = "log:implies";
  static final String N3means = "<http://www.w3.org/2000/10/swap/log#means>";
  static final String N3a = "a";
  static final String N3e = "=";
  static final String RDF = "<http://www.w3.org/1999/02/22-rdf-syntax-ns#>";
  static final String RDFtype = "<http://www.w3.org/1999/02/22-rdf-syntax-ns#type>";
  static final String RDFS = "<http://www.w3.org/2000/01/rdf-schema#>";
  static final String RDFSsubPropertyOf = "<http://www.w3.org/2000/01/rdf-schema#subPropertyOf>";
  static final String DAML = "<http://www.daml.org/2001/03/daml-ont#>";
  static final String DAMLequivalentTo = "<http://www.daml.org/2001/03/daml-ont#equivalentTo>";
  static final String DAMLUniqueProperty = "<http://www.daml.org/2001/03/daml-ont#UniqueProperty>";
  static final String E = "<http://www.agfa.com/w3c/euler#>";
  static final String ENonIrreflexiveProperty = "<http://www.agfa.com/w3c/euler#NonIrreflexiveProperty>";
  static int step = 1;
  static boolean trace = false;
  static boolean core = false;

/**
  constructs a proof engine
                                                                                                    */
  public Euler() {
  }

/**
  asserts all facts and rules acquired from the RDF URI in that proof engine.
  @param uri of the RDF resource
                                                                                                    */
  public synchronized void assert(String uri) {
    String rk = fromWeb(uri);
    String u = baseURI;
    if (asserted == null) asserted = new Vector();
    if (asserted.contains(baseURI)) return;
    Vector vt = new Vector();
    Parser np = new Parser(rk, vt, this, u);
    if (ns == null) ns = new Hashtable();
    if (nsp == null) nsp = new Hashtable();
    if (syn == null) syn = new Hashtable();
    if (anon == null) anon = new Hashtable();
    if (refl == null) refl = new Vector();
    if (uniq == null) uniq = new Vector();
    ns.put("log:", N3);
    while (true) {
      far.near = np.parse();
      if (far.near == null) break;
      while (far.near != null) {
        if (far.near.verb.equals(N3forAll)) {
          if (far.near.subject.subject == null && far.near.subject.object == null)
            vt.addElement(far.near.object.verb);
          else {
            Euler fn = far.near;
            Vector wt = new Vector();
            while (fn != null && fn.verb.equals(N3forAll)) {
              wt.addElement(fn.object.verb);
              fn = fn.near;
            }
            if (wt.size() > varmax) varmax = wt.size();
            far.near.subject.near = fn;
            far.near = far.near.subject;
            far.near.forX(false, wt);
          }
        }
        else if (far.near.verb.equals(RDFSsubPropertyOf)) {
          String s = "data:,{{:s " + far.near.subject.cverb + " :o} log:implies " +
                     "{:s " + far.near.object.cverb + " :o}} a log:Truth; log:forAll :s, :o.";
          assert(s);
        }
        else if (far.near.subject != null && far.near.subject.verb.startsWith("_:")) {
          Euler el = (Euler)anon.get(far.near.subject.verb);
          if (el == null) anon.put(far.near.subject.verb, far.near);
          else {
            if (el.subject == null && el.verb == null && el.object == null) {
              el.verb = far.near.verb;
              el.object = far.near.object;
              el.cverb = far.near.cverb;
              el.varid = far.near.varid;
              el.bound = far.near.bound;
              el.ref = far.near.ref;
              el.near = far.near.near;
            }
            else {
              while (el.near != null) el = el.near;
              el.near = far.near;
            }
          }
          far.near.subject = null;
          far.near = null;
        }
        else if (far.near.object != null && far.near.object.verb.startsWith("_:")) {
          if (anon.get(far.near.object.verb) == null) anon.put(far.near.object.verb, new Euler());
          far.near.object = (Euler)anon.get(far.near.object.verb);
          far = far.near;
        }
        else if (far.near.subject.subject == null && far.near.subject.object != null && !far.near.verb.equals("")) {
          Euler eu = np.parse("");
          eu.subject = np.parse(far.near.cverb);
          eu.subject.object = far.near.object;
          eu.subject.near = far.near.subject;
          eu.object = np.parse("");
          eu.near = far.near.near;
          anon.put("_:" + eu.subject.hashCode(), eu.subject);
          far.near.near = eu;
          if (trace) System.err.println("# " + far.near);
          far = far.near;
        }
        if (far.near != null) {
          if (!core && far.near.verb.equals(N3implies)) far.near.object.ref = far.near.subject;
          if (trace) System.err.println("# " + far.near);
          if (far.near.subject != null) far = far.near;
        }
      }
    }
    for (Enumeration e = anon.elements(); e.hasMoreElements(); ) {
      far.near = np.parse(N3forSomeQN);
      far.near.subject = np.parse("this");
      far.near.object = (Euler)e.nextElement();
      if (trace) System.err.println("# " + far.near);
      far = far.near;
    }
    anon.clear();
    if (vt.size() > varmax) varmax = vt.size();
    asserted.addElement(baseURI);
  }

/**
  proofs a conjunction with that proof engine
  @param uri of the RDF resource
  @return proof
                                                                                                    */
  public synchronized String proof(String uri) {
    String cj = fromWeb(uri);
    String u = baseURI;
    Vector vt = new Vector();
    Parser np = new Parser(cj, vt, this, u);
    if (ns == null) ns = new Hashtable();
    if (nsp == null) nsp = new Hashtable();
    if (syn == null) syn = new Hashtable();
    if (anon == null) anon = new Hashtable();
    if (refl == null) refl = new Vector();
    if (uniq == null) uniq = new Vector();
    ns.put("log:", N3);
    Euler goal = new Euler();
    Euler g = goal;
    while (true) {
      g.near = np.parse();
      if (g.near == null) break;
      while (g.near != null && g.near.verb != null) {
        if (g.near.verb.equals(N3forSome) || g.near.verb.equals(N3forAll) || g.near.verb.equals(N3Truth)) {
          if (g.near.subject.subject == null && g.near.subject.object == null) {
            if (g.near.object != null) vt.addElement(g.near.object.verb);
          }
          else {
            Euler fn = g.near;
            Vector wt = new Vector();
            while (fn != null && (fn.verb.equals(N3forSome) || fn.verb.equals(N3forAll) ||
                   fn.verb.equals(N3Truth))) {
              if (fn.verb.equals(N3forSome) || fn.verb.equals(N3forAll)) wt.addElement(fn.object.verb);
              fn = fn.near;
            }
            if (wt.size() > varmax) varmax = wt.size();
            g.near = g.near.subject;
            Euler gn = g;
            while (gn.near != null) {
              gn.near.forX(false, wt);
              gn = gn.near;
            }
            gn.near = fn;
          }
          g = g.near;
        }
        else if (g.near.subject != null && g.near.subject.verb.startsWith("_:")) {
          Euler el = (Euler)anon.get(g.near.subject.verb);
          if (el == null) anon.put(g.near.subject.verb, g.near);
          else {
            if (el.subject == null && el.verb == null && el.object == null) {
              el.verb = g.near.verb;
              el.object = g.near.object;
              el.cverb = g.near.cverb;
              el.varid = g.near.varid;
              el.bound = g.near.bound;
              el.ref = g.near.ref;
              el.near = g.near.near;
            }
            else {
              while (el.near != null) el = el.near;
              el.near = g.near;
            }
          }
          g.near.subject = null;
          g.near = null;
        }
        else if (g.near.object != null && g.near.object.verb.startsWith("_:")) {
          if (anon.get(g.near.object.verb) == null) anon.put(g.near.object.verb, new Euler());
          g.near.object = (Euler)anon.get(g.near.object.verb);
          g = g.near;
        }
        else if (g.near.subject.subject == null && g.near.subject.object != null && !g.near.verb.equals("")) {
          Euler eu = g.near.subject;
          while (eu.near != null) eu = eu.near;
          eu.near = np.parse(g.near.cverb);
          eu.near.object = g.near.object;
          g.near.verb = "";
          g.near.cverb = "";
          g.near.object = np.parse("");
          anon.put("_:" + g.near.subject.hashCode(), g.near.subject);
          g = g.near;
        }
        else g = g.near;
      }
    }
    for (Enumeration e = anon.elements(); e.hasMoreElements(); ) {
      g.near = np.parse(N3forSomeQN);
      g.near.subject = np.parse("this");
      g.near.object = (Euler)e.nextElement();
      g = g.near;
    }
    anon.clear();
    if (vt.size() > varmax) varmax = vt.size();
    StringBuffer nsp = new StringBuffer(256);
    Enumeration en = ns.keys();
    while (en.hasMoreElements()) {
      Object nsx = en.nextElement();
      nsp.append("@prefix ");
      nsp.append(nsx);
      nsp.append(' ');
      nsp.append(ns.get(nsx));
      nsp.append(".\n");
    }
    nsp.append('\n');
    int tab = 0;
    Euler[] vars = new Euler[varmax];
    Euler lr = goal.near;
    Euler terms = null;
    Euler termsb = null;
    Euler terms1 = null;
    Euler terms2 = null;
    Stack stack = new Stack();
    boolean not = false;
    int istep = step;
    StringBuffer proof = new StringBuffer(4096);
    proof.append("# Generated with Euler ");
    proof.append(version);
    proof.append(" on ");
    proof.append(new Date());
    proof.append('\n');
    proof.append("# for query ");
    proof.append(baseURI);
    proof.append("\n# given ");
    proof.append(asserted);
    proof.append('\n');
    proof.append('\n');
    int pn = proof.length();
    proof.append(nsp);
    int pns = proof.length();
    proof.ensureCapacity(proof.length() * 2);
    StringBuffer pp = new StringBuffer(4096);
    Euler te = goal;
    while (te.near != null) {
      te.near = te.near.eval(vars);
      te = te.near;
    }
    goal = goal.near;
    if (trace) System.err.println("[" + step + "]CALL: " + goal);
    long t = currentTimeMicros();
    while (true) {
      if (goal == null) {
        t = currentTimeMicros() - t + 1;
        if (proof.length() == pns) {
          proof.setLength(pn);
          proof.append("# No proof found for ");
        }
        else proof.append("# Proof found for ");
        proof.append(u);
        proof.append(" in ");
        proof.append(step - istep);
        proof.append(step==1?" step":" steps");
        proof.append(" (");
        proof.append((step - istep)*1000000L/t);
        proof.append(" steps/sec)");
        return proof.toString();
      }
      not = false;
      Euler gl = goal;
      if (goal.varid != -1 && goal.subject != null && goal.object != null) {
        Euler gn = goal.eval(vars);
        goal.verb = gn.verb;
        if (!goal.cverb.equals(N3a) || core) goal.cverb = gn.cverb;
      }
      if (goal.verb == null) {
        if (!core && !refl.contains(goal.object.verb) && goal.object.subject.unify(goal.object.object, this, null)) {
          if (trace) System.err.println("@@ irreflexive " + goal.object);
          not = true;
        }
        tab--;
        if (pp.charAt(pp.length() - 1) == '.') pp.setCharAt(pp.length() - 1, '}');
        else pp.append('}');
        pp.append(' ');
        pp.append(N3impliesQN);
        pp.append('\n');
        for (int i = tab; i > 0; i--) pp.append("  ");
        pp.append('{');
        pp.append(goal.object);
        if (pp.charAt(pp.length() - 1) == '.') pp.setCharAt(pp.length() - 1, '}');
        else pp.append('}');
        pp.append('.');
      }
      else if (goal.verb.equals(N3forAll) || goal.verb.equals(N3forSome) || goal.verb.equals(N3e));
      else if (!core && goal.verb.equals(N3implies) || goal.verb.equals(N3Falsehood)) {
        if (goal.subject.subject != null && goal.subject.object != null) {
          if (goal.subject.varid != -1) {
            Euler gs = goal.subject.ref.eval(vars);
            goal.subject.verb = gs.verb;
            if (!goal.subject.cverb.equals(N3a) || core) goal.subject.cverb = gs.cverb;
          }
          String rk = "";
          Euler el = np.parse(N3Falsehood);
          StringBuffer pv = new StringBuffer(256);
          g = goal.subject;
          while (g != null) {
            Euler gs = g.eval(vars);
            Euler ep = goal.near;
            while (ep != null) {
              if (ep.verb == null
                  && goal.verb.equals(N3Falsehood) && ep.object.unify(gs, this, null)) break;
              ep = ep.near;
            }
            if (ep != null) {
              if (trace) System.err.println("@@ cycle " + ep);
              break;
            }
            rk = proof("data:," + nsp + g.toString());
            vt = new Vector();
            np = new Parser(rk, vt, this, u);
            el = np.parse();
            if ((g.unify(el, this, stack) && !goal.verb.equals(N3Falsehood))
                || goal.object.unify(el, this, stack)) {
              if (pv.length() > 0 && pv.charAt(pv.length() - 1) == '.') {
                pv.append('\n');
                for (int i = tab; i >= 0; i--) pv.append("  ");
              }
              if (goal.verb.equals(N3Falsehood)) pv.append(g);
              else pv.append(el);
            }
            else break;
            g = g.near;
          }
          if (g == null) {
            if (!pv.toString().startsWith(N3Falsehood)) {
              if (pp.length() > 0 && pp.charAt(pp.length() - 1) == '.') {
                pp.append('\n');
                for (int i = tab; i > 0; i--) pp.append("  ");
              }
              pp.append('{');
              pp.append(pv);
              if (pp.charAt(pp.length() - 1) == '.') pp.setCharAt(pp.length() - 1, '}');
              else pp.append('}');
              pp.append(' ');
              pp.append(goal.cverb);
              if (goal.object.subject == null && goal.object.object == null) {
                pp.append(' ');
                pp.append(goal.object);
              }
              else {
                pp.append('\n');
                for (int i = tab; i > 0; i--) pp.append("  ");
                pp.append('{');
                pp.append(goal.object);
                if (pp.charAt(pp.length() - 1) == '.') pp.setCharAt(pp.length() - 1, '}');
                else pp.append('}');
              }
              pp.append('.');
            }
          }
          else not = true;
        }
        else if (goal.object.unify(goal.subject, this, stack) || goal.subject.verb.equals(N3Falsehood)) {
          if (pp.length() > 0 && pp.charAt(pp.length() - 1) == '.') {
            pp.append('\n');
            for (int i = tab; i > 0; i--) pp.append("  ");
          }
          pp.append(goal.subject);
          pp.append(' ');
          pp.append(goal.cverb);
          pp.append(' ');
          if (goal.object.subject != null) pp.append('{');
          pp.append(goal.object);
          if (goal.object.subject != null) {
            if (pp.charAt(pp.length() - 1) == '.') pp.setCharAt(pp.length() - 1, '}');
            else pp.append('}');
          }
          pp.append('.');
        }
        else not = true;
      }
      else {
        if (terms == null) {
          lr = this.near;
          while (lr != null) {
            terms = lr;
            if (terms.verb.equals(goal.verb)) break;
            if (!core && terms.verb.equals(N3implies) && terms.object.verb.equals(goal.verb)) {
              terms = terms.object;
              break;
            }
            else lr = lr.near;
            terms = null;
          }
        }
        if (terms != null) {
          Euler n = lr.near;
          termsb = null;
          while (n != null) {
            termsb = n;
            if (termsb.verb.equals(goal.verb)) break;
            if (!core && termsb.verb.equals(N3implies) && termsb.object.verb.equals(goal.verb)) {
              termsb = termsb.object;
              break;
            }
            else n = n.near;
            termsb = null;
          }
          if (termsb != null) {
            stack.push(goal);
            stack.push(termsb);
            stack.push(n);
            stack.push(new Integer(pp.length()));
            stack.push(new Integer(tab));
          }
          int hc = terms.hashCode();
          for (int i = 0; i < varmax; i++) vars[i] = null;
          if (terms.eval(vars).unify(goal, this, stack)) {
            gl = goal.eval(vars);
            Euler ep = goal.near;
            while (ep != null) {
              if (ep.verb == null && hc == ep.varid && ep.object.unify(gl, this, null)) break;
              ep = ep.near;
            }
            if (ep != null)  {
              if (trace) System.err.println("@@ cycle " + ep);
              not = true;
            }
            else {
              terms = terms.ref;
              if (terms != null) {
                terms2 = new Euler();
                terms1 = terms2;
                while (terms != null) {
                  terms1.near = terms.eval(vars);
                  terms1 = terms1.near;
                  terms = terms.near;
                }
                terms1.near = new Euler();
                terms1 = terms1.near;
                terms1.cverb = "[" + step + "]";
                terms1.object = goal;
                terms1.bound = true;
                terms1.varid = hc;
                terms1.near = goal.near;
                terms1.far = goal.far;
                goal = terms2;
                if (pp.length() > 0 && pp.charAt(pp.length() - 1) == '.') {
                  pp.append('\n');
                  for (int i = tab; i > 0; i--) pp.append("  ");
                }
                if (pp.length() == 0) pp.append(' ');
                else if (pp.length() > 0 && pp.charAt(pp.length() - 1) != '{') pp.append(' ');
                pp.append('{');
                tab++;
              }
              else {
                if (!core && !refl.contains(gl.verb) && gl.subject.unify(gl.object, this, null)) {
                  if (trace) System.err.println("@@ irreflexive " + gl);
                  not = true;
                }
                if (pp.length() > 0 && pp.charAt(pp.length() - 1) == '.') {
                  pp.append('\n');
                  for (int i = tab; i > 0; i--) pp.append("  ");
                }
                pp.append(gl);
              }
            }
          }
          else not = true;
        }
        else not = true;
      }
      if (trace) {
        if (!not && goal.verb != null) System.err.println("[" + step + "]EXIT: " + goal);
        if (!not && gl.verb == null) System.err.println(gl.cverb + "EXIT: " + gl.object);
        if (not && goal.verb != null) System.err.println("[" + step + "]FAIL: " + goal);
        if (not && gl.verb == null) System.err.println(gl.cverb + "FAIL: " + gl.object);
      }
      if (gl.verb != null) step++;
      goal = goal.near;
      if (trace && !not && goal != null && goal.verb != null)
        System.err.println("[" + step + "]CALL: " + goal);
      if (!not && goal == null) {
        proof.append(pp);
        proof.append("\n\n");
      }
      if (not || (!core && goal == null)) {
        goal = null;
        while (!stack.empty()) {
          Object b = stack.pop();
          if (b instanceof Euler) {
            ((Euler)b).bound = false;
            ((Euler)b).ref = null;
          }
          else if (b instanceof Integer) {
            tab = ((Integer)b).intValue();
            pp.setLength(((Integer)stack.pop()).intValue());
            lr = (Euler)stack.pop();
            terms = (Euler)stack.pop();
            goal = (Euler)stack.pop();
            if (trace) System.err.println("[" + step + "]REDO: " + goal);
            break;
          }
        }
      }
    }
  }

  final Euler eval(Euler[] v) {
    if (ref != null && subject == null && object == null) return ref.eval(v);
    if (bound || (varid != -1 && subject != null && object != null)) {
      Euler t = new Euler();
      t.verb = verb;
      t.cverb = cverb;
      if (subject != null) t.subject = subject.eval(v);
      if (object != null) t.object = object.eval(v);
      t.varid = varid;
      t.bound = bound;
      if (v != null && varid != -1 && subject != null && object != null) t.ref = v[varid];
      t.near = near;
      t.far = far;
      if (subject == null && t.subject == null && object != null && t.object != null && near != null)
        t.near = near.eval(v);
      return t;
    }
    if (v != null && v[varid] == null) {
      v[varid] = new Euler();
      v[varid].verb = verb;
      v[varid].cverb = cverb;
      v[varid].varid = varid;
      v[varid].near = near;
      v[varid].far = far;
    }
    if (v != null) return v[varid];
    else return null;
  }

  final boolean unify(Euler t, Euler r, Stack s) {
    if (t == null) return false;
    if (bound && ref != null && subject == null && object == null) return ref.unify(t, r, s);
    if (t.bound && t.ref != null && t.subject == null && t.object == null) return unify(t.ref, r, s);
    if (bound && t.bound || core) {
      if (subject == null && t.subject == null && object != null && t.object != null || core && s == null) {
        Euler el = this;
        Euler elt = t;
        if (s == null) {
          while (el != null && elt != null) {
            el = el.near;
            elt = elt.near;
          }
          if (el != null || elt != null) return false;
          el = this;
          elt = t;
        }
        while (elt != null) {
          el = this;
          while (el != null) {
            if (bound && t.bound && el.verb.equals(elt.verb) || core && el.varid != -1 && el.varid == elt.varid) {
              if (core) {
                if (el.subject != null && !el.subject.unify(elt.subject, r, null) ||
                    el.object != null && !el.object.unify(elt.object, r, null)) {
                  if (r.uniq.contains(el.verb)) return false;
                }
                else break;
              }
              else if (el.object != null && !el.object.unify(elt.object, r, s)) {
                if (r.uniq.contains(el.verb)) return false;
              }
              else break;
            }
            el = el.near;
          }
          if (el == null) return false;
          elt = elt.near;
        }
      }
      else {
        if (!verb.equals(t.verb)) return false;
        if (subject != null && t.subject != null && !subject.unify(t.subject, r, core?null:s)) return false;
        if (object != null && t.object != null && !object.unify(t.object, r, core?null:s)) return false;
      }
      return true;
    }
    if (!core && bound) return t.unify(this, r, s);
    if (s != null) {
      bound = true;
      ref = t;
      s.push(this);
    }
    return true;
  }

  final Euler deref() {
    if (ref != null && subject == null && object == null) return ref.deref();
    else return this;
  }

  final void forX(boolean f, Vector wt) {
    if (subject != null) subject.forX(true, wt);
    if (wt.indexOf(verb) != -1) {
      varid = wt.indexOf(verb);
      bound = false;
    }
    if (object != null) object.forX(true, wt);
    if (f && near != null) near.forX(f, wt);
  }

  final void all(boolean f, Vector v) {
    if (subject != null && verb != null) subject.deref().all(!verb.equals(N3Truth), v);
    if (cverb != null && varid != -1 && !bound && !v.contains(cverb)) v.addElement(cverb);
    if (object != null) object.deref().all(true, v);
    if (f && near != null) near.deref().all(true, v);
  }

  final String string(int n) {
    if (ref != null && subject == null && object == null) return ref.string(0);
    if (subject == null && object == null) return cverb;
    StringBuffer r = new StringBuffer(64);
    Euler el = null;
    if (core && subject != null || verb != null && verb.equals(N3implies) || subject != null && subject.deref().subject == null) {
      el = subject.deref();
      while (el != null) {
        r.append(el.string(n++));
        el = el.near;
      }
    }
    else if (subject != null) r.append(subject.string(n++));
    if (r.length() > 0 && (r.charAt(r.length() - 1) == '.' || r.charAt(r.length() - 1) == ';')) {
      if (r.charAt(0) != ' ') {
        r.insert(0, '{');
        r.setCharAt(r.length() - 1, '}');
      }
      else {
        r.insert(0, '[');
        r.setCharAt(r.length() - 1, ']');
      }
    }
    r.append(' ');
    r.append(cverb);
    if (n % 2 == 0) {
      r.append('\n');
      for (int i = n; i >= 0; i--) r.append("  ");
      n++;
    }
    else r.append(' ');
    int i = r.length();
    if (verb != null && verb.equals(N3implies) || object != null && object.deref().subject == null || core) {
      el = object.deref();
      while (el != null) {
        r.append(el.string(n++));
        el = el.near;
      }
    }
    else if (object != null) r.append(object.string(n++));
    if (r.length() > 0 && (r.charAt(r.length() - 1) == '.' || r.charAt(r.length() - 1) == ';')) {
      if (r.charAt(i) != ' ') {
        r.insert(i, '{');
        r.setCharAt(r.length() - 1, '}');
      }
      else {
        r.insert(i, '[');
        r.setCharAt(r.length() - 1, ']');
      }
    }
    if (subject != null) r.append('.');
    else r.append(';');
    return r.toString();
  }

/**
  string represention of this Euler object
  @return string
                                                                                                    */
  public final String toString() {
    Vector v = new Vector();
    deref().all(false, v);
    if (v.isEmpty() || verb == null) return string(0);
    else {
      StringBuffer sb = new StringBuffer();
      sb.append('{');
      sb.append(string(0));
      if (sb.charAt(sb.length() - 1) == '.') sb.setCharAt(sb.length() - 1, '}');
      else sb.append('}');
      sb.append(" ");
      sb.append(verb.equals(N3implies)||verb.equals(N3Truth)?N3forAllQN:N3forSomeQN);
      sb.append(" ");
      for (Enumeration e = v.elements(); e.hasMoreElements(); ) {
        sb.append(e.nextElement());
        if (e.hasMoreElements()) sb.append(", ");
      }
      sb.append('.');
      return sb.toString();
    }
  }

/**
  raw dump of this and near Euler objects
  @return dump
                                                                                                    */
  public String dump() {
    Euler el = this;
    StringBuffer sb = new StringBuffer();
    int i = 0;
    while (el != null) {
      sb.append('[').append(i++).append("]\n");
      sb.append("subject  ").append(el.subject).append("\n");
      sb.append("object   ").append(el.object).append("\n");
      sb.append("verb     ").append(el.verb).append("\n");
      sb.append("cverb    ").append(el.cverb).append("\n");
      sb.append("varid    ").append(el.varid).append("\n");
      sb.append("bound    ").append(el.bound).append("\n");
      sb.append("ref      ").append(el.ref).append("\n");
      sb.append("near     ").append(el.near).append("\n");
      sb.append("far      ").append(el.far).append("\n");
      sb.append("ns       ").append(el.ns).append("\n");
      sb.append("nsp      ").append(el.nsp).append("\n");
      sb.append("syn      ").append(el.syn).append("\n");
      sb.append("refl     ").append(el.refl).append("\n");
      sb.append("uniq     ").append(el.uniq).append("\n");
      sb.append("varmax   ").append(el.varmax).append("\n");
      sb.append("baseURI  ").append(el.baseURI).append("\n");
      sb.append("asserted ").append(el.asserted).append("\n");
      sb.append("\n");
      el = el.near;
    }
    return sb.toString();
  }

  final String fromWeb(String uri) {
    StringBuffer sb = new StringBuffer();
    String rl = null;
    if (uri.startsWith("data:")) return uri.substring(uri.indexOf(',') + 1);
    else if (uri.startsWith("http:")) {
      try {
        baseURI = uri;
        if (System.getProperty("http_proxy") != null && !System.getProperty("http_proxy").equals(""))
          uri = System.getProperty("http_proxy") + '/' + uri;
        URL r = new URL(uri);
        Socket so = new Socket(r.getHost(), r.getPort()==-1?80:r.getPort());
        so.setTcpNoDelay(true);
        DataOutputStream dos = new DataOutputStream(new BufferedOutputStream(so.getOutputStream()));
        DataInputStream dis = new DataInputStream(new BufferedInputStream(so.getInputStream()));
        PrintWriter pw = new PrintWriter(new OutputStreamWriter(dos, "8859_1"), true);
        pw.println("GET " + r.getFile() + " HTTP/1.0");
        pw.println("Host: " + r.getHost());
        pw.println("");
        pw.flush();
        BufferedReader br = new BufferedReader(new InputStreamReader(dis, "8859_1"));
        while (!br.readLine().equals(""));
        while ((rl = br.readLine()) != null) {
          sb.append(rl);
          sb.append('\n');
        }
        dis.close();
        dos.close();
        so.close();
      }
      catch (Exception e) {
        System.err.println(e);
      }
    }
    else {
      if (uri.startsWith("file:")) uri = uri.substring(uri.indexOf(':') + 1);
      try {
        String path = System.getProperty("user.dir").replace(File.separatorChar, '/');
        path = path.substring(path.indexOf('/')) + "/";
        baseURI = "file:" + (uri.charAt(0)!='/'?path:"") + uri;
        FileInputStream fis = new FileInputStream(uri);
        DataInputStream dis = new DataInputStream(new BufferedInputStream(fis));
        BufferedReader br = new BufferedReader(new InputStreamReader(dis, "8859_1"));
        while ((rl = br.readLine()) != null) {
          sb.append(rl);
          sb.append('\n');
        }
        dis.close();
        fis.close();
      }
      catch (Exception e) {
        System.err.println(e);
      }
    }
    return sb.toString();
  }
  
  static long currentTimeMicros() {
    try {
      System.load("C:\\Program Files\\Portex\\pxtools.dll");
      Class cla = Class.forName("tools.util.Sys");
      return ((Long)cla.getMethod("currentTimeMicros", null).invoke(cla, null)).longValue();
    }
    catch (Throwable t) {
      return System.currentTimeMillis() * 1000L;
    }
  }

/**
  Main method invoked via <code>java Euler</code>
  @param args [-trace] axiom ... lemma
                                                                                                    */
  public static void main(String[] args) {
    if (args.length == 0)
      System.out.println("java [-Dhttp_proxy=%http_proxy%] Euler [-trace] [-core] axiom ... lemma");
    else {
      Euler e = new Euler();
      for (int j = 0; j < args.length - 1; j++) {
        if (args[j].equals("-trace")) trace = true;
        else if (args[j].equals("-core")) core = true;
        else e.assert(args[j]);
      }
      System.out.println(e.proof(args[args.length - 1]));
    }
  }
}
